/*
 * Copyright (c) Contributors to the Open 3D Engine Project. For complete copyright and license terms please see the LICENSE at the root of this distribution.
 * 
 * SPDX-License-Identifier: Apache-2.0 OR MIT
 *
 */

#pragma once

#include <AzCore/Component/EntityId.h>
#include <AzCore/Math/MathUtils.h>
#include <AzCore/Memory/Memory.h>
#include <AzCore/Memory/SystemAllocator.h>
#include <AzCore/RTTI/RTTI.h>
#include <AzCore/RTTI/BehaviorContext.h>
#include <AzCore/RTTI/ReflectContext.h>
#include <AzCore/Serialization/SerializeContext.h>
#include <AzCore/std/hash.h>
#include <AzCore/Component/EntityUtils.h>
#include <AzCore/Component/NamedEntityId.h>

#include <Core/NamedId.h>
#include <ScriptCanvas/Grammar/PrimitivesDeclarations.h>

namespace AZ
{
    class Entity;
    class ReflectContext;
    
    template<typename t_Attribute, typename t_Container>
    bool ReadAttribute(t_Attribute& resultOut, AttributeId id, const t_Container& attributes)
    {
        AZ::Attribute* attribute = FindAttribute(id, attributes);
        return attribute && AZ::AttributeReader(nullptr, attribute).Read<t_Attribute>(resultOut);
    }
}

namespace ScriptCanvas
{
    // A place holder identifier for the AZ::Entity that owns the graph.
    // The actual value in each location initialized to GraphOwnerId is populated with the owning entity at editor-time, Asset Processor-time, or runtime, as soon as the owning entity is known.
    using GraphOwnerIdType = AZ::EntityId;
    static const GraphOwnerIdType GraphOwnerId = AZ::EntityId(0xacedc0de);
    
    // A place holder identifier for unique runtime graph on Entity that is running more than one instance of the same graph.
    // This allows multiple instances of the same graph to be addressed individually on the same entity.
    // The actual value in each location initialized to UniqueId is populated at run-time.
    using RuntimeIdType = AZ::EntityId;
    static const RuntimeIdType UniqueId = AZ::EntityId(0xfee1baad);

    constexpr const char* k_EventOutPrefix = "ExecutionSlot:";

    constexpr const char* k_OnVariableWriteEventName = "OnVariableValueChanged";
    constexpr const char* k_OnVariableWriteEbusName = "VariableNotification";
    
    class Node;
    class Edge;

    using ID = AZ::EntityId;

    using NamespacePath = AZStd::vector<AZStd::string>;
    bool IsNamespacePathEqual(const NamespacePath& lhs, const NamespacePath& rhs);

    using NodeIdList = AZStd::vector<ID>;
    using NodePtrList = AZStd::vector<Node*>;
    using NodePtrConstList = AZStd::vector<const Node*>;

    enum class GrammarVersion : int
    {
        Initial = -1,
        BackendSplit = 0,

        // add new entries above
        Current,
    };

    enum class RuntimeVersion : int
    {
        Initial = -1,
        DirectTraversal = 0,

        // add new entries above
        Current,
    };

    enum class PropertyStatus : AZ::u8
    {
        Getter,
        None,
        Setter,
    };

    struct VersionData
    {
        AZ_TYPE_INFO(VersionData, "{14C629F6-467B-46FE-8B63-48FDFCA42175}");

        static void Reflect(AZ::ReflectContext* context);

        static VersionData Latest();

        GrammarVersion grammarVersion = GrammarVersion::Initial;
        RuntimeVersion runtimeVersion = RuntimeVersion::Initial;

        bool operator == (const VersionData& rhs) const
        {
            return grammarVersion == rhs.grammarVersion && runtimeVersion == rhs.runtimeVersion;
        }

        bool IsLatest() const
        {
            return (*this) == Latest();
        }

        void MarkLatest();
    };

    enum class EventType
    {
        Broadcast,
        BroadcastQueue,
        Event,
        EventQueue,
        Count,
    };

    enum class ExecutionMode : AZ::u8
    {
        Interpreted,
        Native,
        COUNT
    };

    struct SlotId
    {
        AZ_TYPE_INFO(SlotId, "{14C629F6-467B-46FE-8B63-48FDFCA42175}");

        AZ::Uuid m_id{ AZ::Uuid::CreateNull() };

        SlotId() = default;

        SlotId(const SlotId&) = default;

        explicit SlotId(const AZ::Uuid& uniqueId)
            : m_id(uniqueId)
        {}

        //! AZ::Uuid has a constructor not marked as explicit that accepts a const char*
        //! Adding a constructor which accepts a const char* and deleting it prevents
        //! AZ::Uuid from being initialized with c-strings
        explicit SlotId(const char* str) = delete;

        static void Reflect(AZ::ReflectContext* context);

        bool IsValid() const
        {
            return m_id != AZ::Uuid::CreateNull();
        }

        AZStd::string ToString() const
        {
            return m_id.ToString<AZStd::string>();
        }

        bool operator==(const SlotId& rhs) const
        {
            return m_id == rhs.m_id;
        }

        bool operator!=(const SlotId& rhs) const
        {
            return m_id != rhs.m_id;
        }

    };

    struct GraphIdentifier final
    {
        AZ_CLASS_ALLOCATOR(GraphIdentifier, AZ::SystemAllocator, 0);
        AZ_TYPE_INFO(GraphIdentifier, "{0DAFC7EF-D23A-4353-8DA5-7D0CC186D8E3}");

        AZ::ComponentId m_componentId = 0;
        AZ::Data::AssetId m_assetId;

        GraphIdentifier() = default;

        GraphIdentifier(const AZ::Data::AssetId assetId, const AZ::ComponentId& componentId)
            : m_componentId(componentId)
            , m_assetId(assetId)
        {}

        bool operator==(const GraphIdentifier& other) const;

        AZStd::string ToString() const;
    };

    using PropertyFields = AZStd::vector<AZStd::pair<AZStd::string_view, SlotId>>;

    using NamedActiveEntityId = AZ::NamedEntityId;    
    using NamedNodeId = NamedId<AZ::EntityId>;
    using NamedSlotId = NamedId<SlotId>;

    using NodeTypeIdentifier = AZStd::size_t;
    using EBusEventId = AZ::Crc32;
    using EBusBusId = AZ::Crc32;
    using ScriptCanvasId = AZ::EntityId;
    enum class AzEventIdentifier : size_t {};
    
    struct NamespacePathHasher
    {
        AZ_FORCE_INLINE size_t operator()(const NamespacePath& path) const
        {
            return AZStd::hash_range(path.begin(), path.end());
        }
    };

    using DependencySet = AZStd::unordered_set<NamespacePath, NamespacePathHasher>;

    struct DependencyReport
    {
        static DependencyReport NativeLibrary(const AZStd::string& library);

        DependencySet nativeLibraries;
        AZStd::unordered_set<AZ::Data::AssetId> scriptEventsAssetIds;
        DependencySet userSubgraphs;
        AZStd::unordered_set<AZ::Data::AssetId> userSubgraphAssetIds;

        void MergeWith(const DependencyReport& other);
    };

    struct OrderedDependencies
    {
        DependencyReport source;
        AZStd::vector<AZ::Data::AssetId> orderedAssetIds;
    };

    //! Globally accessible Script Canvas settings, we use these to pass user provided settings
    //! into the Script Canvas code
    class ScriptCanvasSettingsRequests : public AZ::EBusTraits
    {
    public:
        static const AZ::EBusAddressPolicy AddressPolicy = AZ::EBusAddressPolicy::Single;

        virtual bool CanShowNetworkSettings() { return false; }

    };

    using ScriptCanvasSettingsRequestBus = AZ::EBus<ScriptCanvasSettingsRequests>;
}

namespace AZStd
{
    template<>
    struct hash<ScriptCanvas::SlotId>
    {
        using argument_type = ScriptCanvas::SlotId;
        using result_type = AZStd::size_t;
        AZ_FORCE_INLINE size_t operator()(const argument_type& ref) const
        {
            return AZStd::hash<AZ::Uuid>()(ref.m_id);
        }
    };
}

#define SCRIPT_CANVAS_INFINITE_LOOP_DETECTION_COUNT (2000000)
